;; Generated package description from swift-mode.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "swift-mode" "10.0.0.0.20260801.20" "Major-mode for Apple's Swift programming language" '((emacs "25.2")) :commit "610b2c24433c11c30a39a6799af69b10cd3bec97" :authors '(("taku0" . "mxxouy6x3m_github@tatapa.org") ("Chris Barrett" . "chris.d.barrett@me.com") ("Bozhidar Batsov" . "bozhidar@batsov.com") ("Arthur Evstifeev" . "lod@pisem.net")) :maintainer '("taku0" . "mxxouy6x3m_github@tatapa.org") :keywords '("languages" "swift") :url "https://github.com/swift-emacs/swift-mode")
